use std::collections::VecDeque;
use std::fmt;

//in numeric representation (quadrants): 1132
const DEFAULT_PASSCODE: [PasscodeItem; 4] = [
    PasscodeItem::TopLeft,
    PasscodeItem::TopLeft,
    PasscodeItem::BottomLeft,
    PasscodeItem::TopRight,
];

#[derive(Copy, Clone, PartialEq, Debug)]
enum PasscodeItem {
    TopLeft,
    TopRight,
    BottomLeft,
    BottomRight,
}

impl TryFrom<char> for PasscodeItem {
    type Error = &'static str;
    fn try_from(value: char) -> Result<Self, Self::Error> {
        match value {
            '1' => Ok(PasscodeItem::TopLeft),
            '2' => Ok(PasscodeItem::TopRight),
            '3' => Ok(PasscodeItem::BottomLeft),
            '4' => Ok(PasscodeItem::BottomRight),
            _ => Err("invalid quadrant for passcodeitem: use 1 (topleft), 2 (topright), 3 (bottomleft) ,4 (bottomright)")
        }
    }
}

impl fmt::Display for PasscodeItem {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        match self {
            PasscodeItem::TopLeft => write!(f, "TOP LEFT"),
            PasscodeItem::TopRight => write!(f, "TOP RIGHT"),
            PasscodeItem::BottomLeft => write!(f, "BOTTOM LEFT"),
            PasscodeItem::BottomRight => write!(f, "BOTTOM RIGHT"),
        }
    }
}

#[derive(Clone, PartialEq, Debug)]
pub struct TouchState {
    passcode: Vec<PasscodeItem>,
    touches: VecDeque<PasscodeItem>,
    max_touches: usize,
    half_width: f64,
    half_height: f64,
}

impl TouchState {
    pub fn new(width: i32, height: i32, passcode: Option<&str>) -> Self {
        Self {
            passcode: if let Some(passcode) = passcode {
                passcode
                    .chars()
                    .map(|x| x.try_into().expect("Invalid passcode"))
                    .collect()
            } else {
                DEFAULT_PASSCODE.iter().copied().collect()
            },
            touches: VecDeque::new(),
            max_touches: if let Some(passcode) = passcode {
                passcode.chars().count()
            } else {
                DEFAULT_PASSCODE.len()
            },
            half_width: width as f64 / 2.0,
            half_height: height as f64 / 2.0,
        }
    }

    pub fn add_touch(&mut self, x_val: f64, y_val: f64) -> bool {
        let touch: Option<PasscodeItem> = if x_val < self.half_width && y_val < self.half_height {
            Some(PasscodeItem::TopLeft)
        } else if x_val > self.half_width && y_val < self.half_height {
            Some(PasscodeItem::TopRight)
        } else if x_val < self.half_width && y_val > self.half_height {
            Some(PasscodeItem::BottomLeft)
        } else if x_val > self.half_width && y_val > self.half_height {
            Some(PasscodeItem::BottomRight)
        } else {
            None
        };

        if let Some(touch) = touch {
            self.touches.push_back(touch);
            while self.touches.len() > self.max_touches {
                self.touches.pop_front();
            }
            true
        } else {
            eprintln!("unknown touch!");
            false
        }
    }

    pub fn print(&self) {
        for touch in self.touches.iter() {
            println!("{}", touch);
        }
    }

    pub fn is_pass_valid(&self) -> bool {
        self.touches == self.passcode
    }
}
