use std::{fs::File, os::fd::AsFd};

use wayland_client::{
    delegate_noop,
    protocol::{
        wl_compositor,
        wl_output::{self, WlOutput},
        wl_registry, wl_seat, wl_surface,
        wl_touch::{self, WlTouch},
    },
    Connection, Dispatch, EventQueue, QueueHandle,
};

use wayland_client::protocol::wl_seat::WlSeat;
use wayland_protocols::ext::session_lock::v1::client::{
    ext_session_lock_manager_v1::{self, ExtSessionLockManagerV1},
    ext_session_lock_surface_v1, ext_session_lock_v1,
};
use wayland_protocols::{
    ext::session_lock::v1::client::ext_session_lock_v1::ExtSessionLockV1,
    xdg::xdg_output::zv1::client::{
        zxdg_output_manager_v1::ZxdgOutputManagerV1,
        zxdg_output_v1::{self, ZxdgOutputV1},
    },
};

use wayland_protocols::xdg::shell::client::{xdg_toplevel, xdg_wm_base};
use wayland_protocols::xdg::xdg_output::zv1::client::zxdg_output_manager_v1;

use wayland_client::protocol::wl_buffer;
use wayland_client::protocol::wl_shm;
use wayland_client::protocol::wl_shm_pool;

#[derive(Copy, Clone)]
pub struct TouchDown {
    pub serial: u32,
    pub x: f64,
    pub y: f64,
}

// This struct represents the state of our app. This simple app does not
// need any state, by this type still supports the `Dispatch` implementations.
pub struct State {
    lock: Option<ExtSessionLockV1>,
    lock_manager: Option<ExtSessionLockManagerV1>,
    connection: Connection,
    queue_handle: QueueHandle<State>,
    compositor: Option<wl_compositor::WlCompositor>,
    output: Option<WlOutput>,
    touch: Option<WlTouch>,
    pub last_down: Option<TouchDown>,
    pub width: Option<i32>,
    pub height: Option<i32>,
    seat: Option<WlSeat>,
    buffer: Option<wl_buffer::WlBuffer>,
    base_surface: Option<wl_surface::WlSurface>,
    flipped: bool,
    configured: bool,
    wm_base: Option<xdg_wm_base::XdgWmBase>,
    zxdg_manager: Option<zxdg_output_manager_v1::ZxdgOutputManagerV1>,
    wl_shm: Option<wl_shm::WlShm>,
}

impl State {
    fn alloc(&mut self) {
        let (init_w, init_h) = (self.width.unwrap(), self.height.unwrap());

        let mut file = tempfile::tempfile().unwrap();
        draw(
            &mut file,
            (init_w.try_into().unwrap(), init_h.try_into().unwrap()),
        );
        let pool = self.wl_shm.as_ref().unwrap().create_pool(
            file.as_fd(),
            (init_w * init_h * 4) as i32,
            &self.queue_handle,
            (),
        );
        let buffer = pool.create_buffer(
            0,
            init_w as i32,
            init_h as i32,
            (init_w * 4) as i32,
            wl_shm::Format::Argb8888,
            &self.queue_handle,
            (),
        );
        self.buffer = Some(buffer.clone());

        if self.configured {
            let surface = self.base_surface.as_ref().unwrap();
            surface.attach(Some(&buffer), 0, 0);
            surface.commit();
        }
    }

    pub fn flip(&mut self) {
        if self.configured {
            let surface = self.base_surface.as_ref().unwrap();
            self.flipped = !self.flipped;
            surface.set_buffer_transform(if self.flipped {
                wl_output::Transform::Flipped180
            } else {
                wl_output::Transform::Normal
            });
            surface.damage_buffer(0, 0, self.width.unwrap(), self.height.unwrap());
            surface.commit();
        }
        self.connection.roundtrip().expect("Failed to flip");
    }

    pub fn lock(&mut self) {
        self.alloc();
        let lock = self
            .lock_manager
            .as_ref()
            .unwrap()
            .lock(&self.queue_handle, ());

        self.lock = Some(lock);
        self.connection.roundtrip().expect("Failed to lock");
    }

    pub fn unlock(&mut self) {
        self.lock.as_ref().unwrap().unlock_and_destroy();
        self.lock = None;
        self.connection.roundtrip().expect("Failed to unlock");
        self.lock_manager.as_ref().unwrap().destroy();
        self.lock_manager = None;
    }

    pub fn round(&self) {
        self.connection.roundtrip().expect("Failed to round");
    }
}

// Implement `Dispatch<WlRegistry, ()> for out state. This provides the logic
// to be able to process events for the wl_registry interface.
//
// The second type parameter is the user-data of our implementation. It is a
// mechanism that allows you to associate a value to each particular Wayland
// object, and allow different dispatching logic depending on the type of the
// associated value.
//
// In this example, we just use () as we don't have any value to associate. See
// the `Dispatch` documentation for more details about this.
impl Dispatch<wl_registry::WlRegistry, ()> for State {
    fn event(
        state: &mut Self,
        registry: &wl_registry::WlRegistry,
        event: wl_registry::Event,
        _: &(),
        _: &Connection,
        qh: &QueueHandle<Self>,
    ) {
        // When receiving events from the wl_registry, we are only interested in the
        // `global` event, which signals a new available global.
        // When receiving this event, we just print its characteristics in this example.
        if let wl_registry::Event::Global {
            name,
            interface,
            version,
        } = event
        {
            eprintln!("[{}] {} (v{})", name, interface, version);
            match &interface[..] {
                "ext_session_lock_manager_v1" => {
                    let lock_manager =
                        registry.bind::<ExtSessionLockManagerV1, _, _>(name, 1, qh, ());
                    state.lock_manager = Some(lock_manager);
                }
                "wl_compositor" => {
                    let compositor =
                        registry.bind::<wl_compositor::WlCompositor, _, _>(name, 4, qh, ());
                    let surface = compositor.create_surface(qh, ());
                    state.base_surface = Some(surface);
                    state.compositor = Some(compositor);
                }
                "wl_output" => {
                    let output = registry.bind::<WlOutput, _, _>(name, 1, qh, ());
                    // Put first output as pin entry output
                    // TODO: allow user to set env variable to choose display
                    // Moreover, this code will cause bugs when new displays are plugged in
                    // Do we just want to display the pin entry on all displays? If so,
                    // use a vector
                    state.output = Some(output);
                }
                "wl_seat" => {
                    let seat = registry.bind::<WlSeat, _, _>(name, 1, qh, ());
                    // TODO: JUST SUPPORT TOUCH! DONT DO cursor https://github.com/Smithay/wayland-rs/issues/85
                    state.touch = Some(seat.get_touch(qh, ()));
                    state.seat = Some(seat);
                }
                "wl_shm" => {
                    let shm = registry.bind::<wl_shm::WlShm, _, _>(name, 1, qh, ());

                    state.wl_shm = Some(shm);
                }
                "xdg_wm_base" => {
                    let wm_base = registry.bind::<xdg_wm_base::XdgWmBase, _, _>(name, 1, qh, ());
                    state.wm_base = Some(wm_base);
                }
                "zxdg_output_manager_v1" => {
                    let zxdg = registry.bind::<zxdg_output_manager_v1::ZxdgOutputManagerV1, _, _>(
                        name,
                        1,
                        qh,
                        (),
                    );
                    state.zxdg_manager = Some(zxdg);
                }
                _ => {}
            }
        }
    }
}

impl Dispatch<wl_touch::WlTouch, ()> for State {
    fn event(
        state: &mut Self,
        touch: &wl_touch::WlTouch,
        event: wl_touch::Event,
        _: &(),
        _: &Connection,
        qh: &QueueHandle<Self>,
    ) {
        match event {
            wl_touch::Event::Down { serial, x, y, .. } => {
                state.last_down = Some(TouchDown {
                    serial: serial,
                    x: x,
                    y: y,
                })
            }
            _ => {}
        }
    }
}

impl Dispatch<ext_session_lock_surface_v1::ExtSessionLockSurfaceV1, ()> for State {
    fn event(
        state: &mut Self,
        lock_surface: &ext_session_lock_surface_v1::ExtSessionLockSurfaceV1,
        event: ext_session_lock_surface_v1::Event,
        _: &(),
        _: &Connection,
        qh: &QueueHandle<Self>,
    ) {
        match event {
            ext_session_lock_surface_v1::Event::Configure {
                serial,
                width,
                height,
            } => {
                state.configured = true;
                lock_surface.ack_configure(serial);
                state.base_surface.as_ref().unwrap().commit();
            }
            _ => todo!(),
        }
    }
}

impl Dispatch<ext_session_lock_v1::ExtSessionLockV1, ()> for State {
    fn event(
        state: &mut Self,
        compositor: &ext_session_lock_v1::ExtSessionLockV1,
        event: ext_session_lock_v1::Event,
        _: &(),
        _: &Connection,
        qh: &QueueHandle<Self>,
    ) {
        match event {
            ext_session_lock_v1::Event::Locked => {
                let surface = state.compositor.as_ref().unwrap().create_surface(&qh, ());
                let lock_surface = state.lock.as_ref().unwrap().get_lock_surface(
                    &surface,
                    state.output.as_ref().unwrap(),
                    &qh,
                    (),
                );
                surface.attach(Some(&state.buffer.clone().unwrap()), 0, 0);
                state.base_surface = Some(surface);
            }
            _ => todo!(),
        }
    }
}

pub fn init_wayland() -> (State, EventQueue<State>) {
    // Create a Wayland connection by connecting to the server through the
    // environment-provided configuration.
    let conn = Connection::connect_to_env().unwrap();

    // Retrieve the WlDisplay Wayland object from the connection. This object is
    // the starting point of any Wayland program, from which all other objects will
    // be created.
    let display = conn.display();

    // Create an event queue for our event processing
    let mut event_queue = conn.new_event_queue();
    // An get its handle to associated new objects to it
    let qh = event_queue.handle();

    // Create a wl_registry object by sending the wl_display.get_registry request
    // This method takes two arguments: a handle to the queue the newly created
    // wl_registry will be assigned to, and the user-data that should be associated
    // with this registry (here it is () as we don't need user-data).
    let _registry = display.get_registry(&qh, ());

    // At this point everything is ready, and we just need to wait to receive the events
    // from the wl_registry, our callback will print the advertized globals.
    eprintln!("Advertized globals:");

    let mut state = State {
        lock: None,
        lock_manager: None,
        connection: conn,
        queue_handle: qh,
        compositor: None,
        output: None,
        touch: None,
        seat: None,
        buffer: None,
        base_surface: None,
        configured: false,
        flipped: false,
        wm_base: None,
        last_down: None,
        width: None,
        height: None,
        zxdg_manager: None,
        wl_shm: None,
    };

    // To actually receive the events, we invoke the `sync_roundtrip` method. This method
    // is special and you will generally only invoke it during the setup of your program:
    // it will block until the server has received and processed all the messages you've
    // sent up to now.
    //
    // In our case, that means it'll block until the server has received our
    // wl_display.get_registry request, and as a reaction has sent us a batch of
    // wl_registry.global events.
    //
    // `sync_roundtrip` will then empty the internal buffer of the queue it has been invoked
    // on, and thus invoke our `Dispatch` implementation that prints the list of advertized
    // globals.
    event_queue.blocking_dispatch(&mut state).unwrap();

    let output = state.zxdg_manager.as_ref().unwrap().get_xdg_output(
        state.output.as_ref().unwrap(),
        &state.queue_handle,
        (),
    ); // launch zxdg dispatch for setting width and height
    event_queue.blocking_dispatch(&mut state).unwrap();

    return (state, event_queue);
}

fn draw(tmp: &mut File, (buf_x, buf_y): (u32, u32)) {
    use std::{cmp::min, io::Write};
    let mut buf = std::io::BufWriter::new(tmp);
    for y in 0..buf_y {
        for x in 0..buf_x {
            let a = 0xFF;
            // let r = min(((buf_x - x) * 0xFF) / buf_x, ((buf_y - y) * 0xFF) / buf_y);
            let r = 0x00;
            let g = min((x * 0xFF) / buf_x, ((buf_y - y) * 0xFF) / buf_y);
            let b = min(((buf_x - x) * 0xFF) / buf_x, (y * 0xFF) / buf_y);

            let color = (a << 24) + (r << 16) + (g << 8) + b;
            buf.write_all(&color.to_ne_bytes()).unwrap();
        }
    }
    buf.flush().unwrap();
}

impl Dispatch<xdg_wm_base::XdgWmBase, ()> for State {
    fn event(
        _: &mut Self,
        wm_base: &xdg_wm_base::XdgWmBase,
        event: xdg_wm_base::Event,
        _: &(),
        _: &Connection,
        _: &QueueHandle<Self>,
    ) {
        if let xdg_wm_base::Event::Ping { serial } = event {
            wm_base.pong(serial);
        }
    }
}

impl Dispatch<xdg_toplevel::XdgToplevel, ()> for State {
    fn event(
        state: &mut Self,
        _: &xdg_toplevel::XdgToplevel,
        event: xdg_toplevel::Event,
        _: &(),
        _: &Connection,
        _: &QueueHandle<Self>,
    ) {
        // if let xdg_toplevel::Event::Close {} = event {
        //     state.running = false;
        // }
    }
}

impl Dispatch<zxdg_output_v1::ZxdgOutputV1, ()> for State {
    fn event(
        state: &mut Self,
        zxdg_output: &zxdg_output_v1::ZxdgOutputV1,
        event: zxdg_output_v1::Event,
        _: &(),
        _: &Connection,
        _: &QueueHandle<Self>,
    ) {
        match event {
            zxdg_output_v1::Event::LogicalSize { width, height } => {
                state.width = Some(width);
                state.height = Some(height);
            }
            _ => {}
        }
    }
}

delegate_noop!(State: ignore wl_shm::WlShm);
delegate_noop!(State: ignore wl_shm_pool::WlShmPool);
delegate_noop!(State: ignore wl_buffer::WlBuffer);
delegate_noop!(State: ignore wl_output::WlOutput);
delegate_noop!(State: ignore ext_session_lock_manager_v1::ExtSessionLockManagerV1);
delegate_noop!(State: ignore wl_surface::WlSurface);
delegate_noop!(State: ignore wl_seat::WlSeat);
delegate_noop!(State: ignore wl_compositor::WlCompositor);
delegate_noop!(State: ignore zxdg_output_manager_v1::ZxdgOutputManagerV1);
